/* Simple, console-style game menu system

Copyright (C) 2009 Damian Yerrick

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "drawing.h"
#include "bfmenu.h"
#include "joyconfig.h"

enum {
  MAX_VIS_ITEMS = 7,
  ITEM_HT = 48,
  MENU_TOP = 48,
  MENU_LEFT = 48,
  MENU_INDENT = 32,
  MAX_FADELEVEL = 10,
  VKEY_PREV = VKEY_MENUUP,
  VKEY_NEXT = VKEY_MENUDOWN,
  // VKEY_MENUOK = VKEY_PREV | VKEY_NEXT,
};

static void drawMenuTitle(const FKMenu *m, int color) {
  textout_ex(screen, picken_sans40, m->title,
             MENU_LEFT, MENU_TOP,
             color, 0);
}

static int getFadedColor(int fadeLevel) {
  return makecol(14 * fadeLevel, 14 * fadeLevel, 14 * fadeLevel);
}

static void drawInstruction(void) {
  textout_ex(screen, picken_sans, "Up or Down: move; OK: select",
             MENU_LEFT, MENU_TOP + (MAX_VIS_ITEMS + 1) * ITEM_HT,
             makecol(255, 255, 255), 0);
}

static void drawMenuItem(const FKMenuItem *item, unsigned int y, int color) {
  textout_ex(screen, picken_sans40, item->name,
             MENU_LEFT + MENU_INDENT, MENU_TOP + ITEM_HT * (y + 1),
             color, 0);
}

/**
 * @param fadeLevel brightness of unselected items, from 0 to 10
 */
static void drawMenuItems(const FKMenuState *m, unsigned int fadeLevel) {
  if (m->menu->numItems <= MAX_VIS_ITEMS) {
    for (unsigned int i = 0; i < m->menu->numItems; ++i) {
      int color = (i == m->selected)
                  ? makecol(255, 255, 153)
                  : getFadedColor(fadeLevel);
      drawMenuItem(&(m->menu->items[i]), i, color);
    }
  } else {
    for (unsigned int i = 0; i < MAX_VIS_ITEMS; ++i) {
      int item = m->selected + m->menu->numItems - (MAX_VIS_ITEMS / 2);

      if (item >= m->menu->numItems) {
        item -= m->menu->numItems;
      }

      int color = (i == m->selected)
                  ? makecol(255, 255, 153)
                  : getFadedColor(fadeLevel);
      drawMenuItem(&(m->menu->items[item]), i, color);
    }
  }
}

static void transitionEnterMenu(FKMenuState *m) {
  if (m->transitionTime == 0) {
    drawMenuItems(m, MAX_FADELEVEL);
    m->transition = 0;
    drawInstruction();
    return;
  }
  if (m->transitionTime >= MAX_FADELEVEL) {
    rectfill(screen,
             0, 48,
             SCREEN_W - 1, MENU_TOP + ITEM_HT * (MAX_VIS_ITEMS + 1) - 1,
             0);
    drawMenuTitle(m->menu, getFadedColor(MAX_FADELEVEL * 3 / 2));
  } else {
    int gray = MAX_FADELEVEL - m->transitionTime;
    drawMenuItems(m, gray);
  }
  --m->transitionTime;
}


/**
 * @param phase 0: fully title; 10: fully item
 */
static void eraseTitleBecomingItem(const FKMenuState *m, unsigned int phase) {
  int x = MENU_LEFT + MENU_INDENT * phase / 10;
  int line = (m->menu->numItems > MAX_VIS_ITEMS
              ? MAX_VIS_ITEMS / 2
              : m->selected) + 1;
  int y = MENU_TOP + ITEM_HT * line * phase / 10;
  int width = text_length(picken_sans40, m->menu->items[m->selected].name);
  rectfill(screen, x, y, x + width - 1, y + ITEM_HT - 1, 0);
}

static void drawTitleBecomingItem(const FKMenuState *m, unsigned int phase) {
  int x = MENU_LEFT + MENU_INDENT * phase / 10;
  int line = (m->menu->numItems > MAX_VIS_ITEMS ? 3 : m->selected) + 1;
  int y = MENU_TOP + ITEM_HT * line * phase / 10;
  textout_ex(screen, picken_sans40, m->menu->items[m->selected].name,
             x, y, makecol(255, 255, 255), 0);
}

static bool transitionOutEnterMenu(FKMenuState *m) {

  // if being set up for the first time, set length of transition
  if (m->transitionTime == 0) {
    m->transitionTime = 10 + MAX_FADELEVEL;
  }

  --m->transitionTime;
  if (m->transitionTime > 10) {

    // Step 1: fade out title and other items
    int fadeLevel = m->transitionTime - 10;
    int titleFadeLevel = fadeLevel * 3 / 2;

    // Fade other items
    drawMenuItems(m, fadeLevel);
    drawMenuTitle(m->menu, getFadedColor(titleFadeLevel));
    return false;
  } else if (m->transitionTime > 0) {
    if (m->transitionTime == 10) {
      fkPlaySounds(FKSND_TOSS);  // menu open
    }

    // Step 2: Move selected item toward menu title position
    eraseTitleBecomingItem(m, m->transitionTime);
    drawTitleBecomingItem(m, m->transitionTime - 1);
    return false;
  } else {

    // Done.
    m->transition = TRANSITION_ENTER_SUBMENU;
    m->transitionTime = 10;
    return true;
  }
}

static bool transitionOutExitMenu(FKMenuState *m) {

  // if being set up for the first time, set length of transition
  if (m->transitionTime == 0) {
    m->transitionTime = MAX_FADELEVEL;
  }

  --m->transitionTime;
  if (m->transitionTime > 0) {

    // Step 1: fade out title and other items
    int fadeLevel = m->transitionTime;

    // Fade other items
    drawMenuItems(m, fadeLevel);
    return false;
  } else {

    // Done.
    m->transition = TRANSITION_EXIT_SUBMENU;
    m->transitionTime = MAX_FADELEVEL + 10;
    return true;
  }
}

static void transitionExitMenu(FKMenuState *m) {
  if (m->transitionTime == 0) {
    drawMenuItems(m, MAX_FADELEVEL);
    m->transition = 0;
    drawInstruction();
    return;
  }
  if (m->transitionTime == MAX_FADELEVEL + 10) {
    fkPlaySounds(FKSND_TOSS);  // menu close
  }

  if (m->transitionTime >= MAX_FADELEVEL + 10) {

    // Step 1: Clear out old menu entirely
    rectfill(screen,
             0, 48,
             SCREEN_W - 1, MENU_TOP + ITEM_HT * (MAX_VIS_ITEMS + 1) - 1,
             0);
    drawTitleBecomingItem(m, 0);
  } else if (m->transitionTime > MAX_FADELEVEL) {

    // Step 2: Move selected item away from menu title position
    int phase = (MAX_FADELEVEL + 10) - m->transitionTime;

    eraseTitleBecomingItem(m, phase - 1);
    drawTitleBecomingItem(m, phase);
  } else {
    int fadeLevel = MAX_FADELEVEL + 1 - m->transitionTime;
    int titleFadeLevel = fadeLevel * 3 / 2;

    if (m->transitionTime == MAX_FADELEVEL) {
      eraseTitleBecomingItem(m, 9);
    }

    int c = getFadedColor(titleFadeLevel);
    drawMenuTitle(m->menu, c);
    drawMenuItems(m, fadeLevel);
  }
  --m->transitionTime;
}

bool fkMenuRun(FKMenuState *m, FKBits vkeys) {
  FKBits downKeys = vkeys & ~m->lastVkeys;
  FKBits upKeys = ~vkeys & m->lastVkeys;
  m->lastVkeys = vkeys;

  if (m->transition == 0) {
    bool dirty = 0;

    // pressed left and right
    if (((vkeys & VKEY_MENUOK) == VKEY_MENUOK)
        && (downKeys & VKEY_MENUOK) != 0) {
      int transition = m->menu->items[m->selected].transition;

      if (transition == 0) {
        return true;
      }
      m->transition = TRANSITION_OUT | transition;
      return false;
    }
    if (upKeys & VKEY_PREV) {
      if (m->selected > 0) {
        --m->selected;
      } else {
        m->selected = m->menu->numItems - 1;
      }
      fkPlaySounds(FKSND_SHIFT);
      dirty = true;
    }
    if (upKeys & VKEY_NEXT) {
      if (m->selected < m->menu->numItems - 1) {
        ++m->selected;
      } else {
        m->selected = 0;
      }
      fkPlaySounds(FKSND_SHIFT);
      dirty = true;
    }
    if (dirty) {
      drawMenuItems(m, 10);
    }
    return false;
  }
  if (m->transition == (TRANSITION_ENTER_SUBMENU | TRANSITION_OUT)) {
    return transitionOutEnterMenu(m);
  }
  if (m->transition == TRANSITION_ENTER_SUBMENU) {
    transitionEnterMenu(m);
    return false;
  }
  if (m->transition == (TRANSITION_EXIT_SUBMENU | TRANSITION_OUT)) {
    return transitionOutExitMenu(m);
  }
  if (m->transition == TRANSITION_EXIT_SUBMENU) {
    transitionExitMenu(m);
    return false;
  }
  return false;
}
