/* FK: tile matching game

Copyright (C) 2009 Damian Yerrick

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <allegro.h>
#include "drawing.h"

#define USE_DEBUG_SET_COLOR true

BITMAP *back;  // the buffer into which the game objects are composited
BITMAP *bg;  // the skin background with some of the borders composited on top of it
PALETTE tile_palette;
DATAFILE *fk_dat = NULL;
DATAFILE *sound_dat = NULL;
const FONT *picken_sans = NULL, *picken_sans40 = NULL, *led_font = NULL;
BITMAP *peacedes, *top_tiles, *top_tiles_flip;
BITMAP *dtot_temp;



int openWindow(bool windowed) {
  int cd = desktop_color_depth();
  int driver = windowed ? GFX_AUTODETECT_WINDOWED : GFX_AUTODETECT_FULLSCREEN;

  // don't use indexed color
  if (cd < 15) {
    cd = 16;
  }
  set_color_depth(cd);
  if (set_gfx_mode(driver, 640, 480, 0, 0) >= 0) {
    return 0;
  }

  // On Windows, desktop_color_depth() doesn't reliably distinguish
  // 15-bit from 16-bit color, so try the other.
  if (cd == 15 || cd == 16) {
    set_color_depth(31 - cd);
    if (set_gfx_mode(driver, 640, 480, 0, 0) >= 0) {
      return 0;
    }
  }
  return -1;
}

const char trackColors[7][3] = {
  // the conveyor front: green, red, yellow, blue, orange
  // the conveyor belt: gray, dark gray
  {0, 3, 0}, {3, 0, 0}, {3, 3, 0}, {0, 1, 3}, {3, 2, 0},
  {2, 2, 2}, {1, 1, 1}
};

int trackpal[7][8];
void setupTrackpal(void) {
  static const unsigned char blacklevel[8] = {96, 64, 24, 23, 22, 21, 16, 16};
  static const unsigned char gain[8] = {50, 50, 50, 49, 48, 47, 40, 30};
  // 0-7: colors for even checkers on conveyor
  // 8-15: colors for odd checkers on conveyor
  for (int fret = 0; fret < 7; ++fret) {
    for (int z = 0; z < 8; z += 1) {
      int r = blacklevel[z] + gain[z] * trackColors[fret][0];
      int g = blacklevel[z] + gain[z] * trackColors[fret][1];
      int b = blacklevel[z] + gain[z] * trackColors[fret][2];
      int c = makecol(r, g, b);
      trackpal[fret][z] = c;
    }
  }
}

const void *fdo(const DATAFILE *dat, const char *name) {
  if (!dat) return NULL;
  DATAFILE *obj = find_datafile_object(dat, name);
  if (!obj) return NULL;
  return obj->dat;
}

bool fkLoadDatafiles(void) {
  LAZYINIT(back, create_bitmap(SCREEN_W, SCREEN_H))

  // draw copyright notice on screen while loading
  int coprNoticeColor = makecol(255, 255, 85);
  clear_to_color(back, makecol(0, 0, 170));
  for (int i = 0; coprNotice[i]; ++i) {
    textout_ex(back, font, coprNotice[i],
               24, 20 + i * 10, coprNoticeColor, -1);
  }
  stretch_blit(back, screen,
               0, 0, SCREEN_W / 2, SCREEN_H / 2,
               0, 0, SCREEN_W, SCREEN_H);
  int loadingTime = ticks + 180;

  set_color_conversion(COLORCONV_MOST);
  LAZYINIT(fk_dat, load_datafile("fk.dat"))
  LAZYINIT(sound_dat, load_datafile("sound.dat"))

  picken_sans = fdo(fk_dat, "picken_sans");
  picken_sans40 = fdo(fk_dat, "picken_sans40");
  led_font = fdo(fk_dat, "led");
  peacedes = (BITMAP *)fdo(fk_dat, "peacedes");
  top_tiles = (BITMAP *)fdo(fk_dat, "top_tiles");
  if (!led_font || !picken_sans || !picken_sans40 || !peacedes
      || !top_tiles) {
    return FALSE;
  }

  LAZYINIT(bg, create_bitmap(SCREEN_W, SCREEN_H))
  LAZYINIT(dtot_temp, create_bitmap(64, 40))
  LAZYINIT(top_tiles_flip, create_bitmap_ex(bitmap_color_depth(top_tiles),
                                            top_tiles->w, top_tiles->h))

  clear_to_color(top_tiles_flip, bitmap_mask_color(top_tiles));
  draw_sprite_h_flip(top_tiles_flip, top_tiles, 0, 0);
  setupTrackpal();

  // if we finished early, we still need to keep the copr notice on screen
  if (ticks < loadingTime) {
    rest(17 * (loadingTime - ticks));
  }
  return TRUE;  
}

void drawing_cleanup(void) {
  DISPOSE(unload_datafile, fk_dat)
  DISPOSE(unload_datafile, sound_dat)
  DISPOSE(destroy_bitmap, back)
  DISPOSE(destroy_bitmap, bg)
  DISPOSE(destroy_bitmap, top_tiles_flip)
  DISPOSE(destroy_bitmap, dtot_temp)
}


#define TRACK_FRONT_WIDTH 256
#define TRACK_HEIGHT 160
#define TRACK_FRONT_Y 224
#define BIN_TOTAL_HT 192
#define BIN_WIDTH 272
#define Z_WHOLES_PER_BARLINE 2
/**
 * Draws the track background.  Use this to add the track to the
 * skin background.
 */
void drawTrack(BITMAP *dst, int xcenter, bool lefty, int zdelta) {

  // the conveyor
  for (int y = TRACK_HEIGHT; y < TRACK_HEIGHT * 2; ++y) {
    unsigned int z = (TRACK_HEIGHT * 2 * TRACK_DEPTH * 8)/y - (TRACK_DEPTH*8);
    unsigned int zsub = (z + zdelta) / Z_WHOLES_PER_BARLINE & 0x07;
    int x1 = xcenter - y * TRACK_FRONT_WIDTH / (TRACK_HEIGHT * 4);
    int x2 = xcenter + y * TRACK_FRONT_WIDTH / (TRACK_HEIGHT * 4);

    for (int x = 0; x < 5; ++x) {
      int fret = 5 + (x & 1);
      int l = x1 + (x2 - x1) * x / N_TRACKS;
      int r = x1 + (x2 - x1) * (x + 1) / N_TRACKS - 1;
      hline(dst, l, y + TRACK_FRONT_Y - 2 * TRACK_HEIGHT, r, trackpal[fret][zsub]);
    }
  }
}

void fkDrawBin(BITMAP *dst, int xcenter, bool lefty) {
  int x1 = xcenter - TRACK_FRONT_WIDTH / 2;
  int x2 = xcenter + TRACK_FRONT_WIDTH / 2;
  
  // the front of the track
  for (int x = 0; x < N_TRACKS; ++x) {
    int l = x1 + (x2 - x1) * x / N_TRACKS;
    int r = x1 + (x2 - x1) * (x + 1) / N_TRACKS - 1;
    
    int fret = lefty ? (N_TRACKS - 1) - x : x;
    for (int y = 0; y < 8; ++y) {
      int c = trackpal[fret][y];
      rectfill(dst, l, TRACK_FRONT_Y + y * 2, r, TRACK_FRONT_Y + y * 2 + 1, c);
    }
  }

  // indicator area below the track
  rectfill(dst,
           xcenter - TRACK_FRONT_WIDTH / 2, TRACK_FRONT_Y + 16,
           xcenter + TRACK_FRONT_WIDTH / 2 - 1, TRACK_FRONT_Y + BIN_TOTAL_HT - 81,
           makecol(0, 0, 0));
  // back of the bin
  rectfill(dst,
           xcenter - BIN_WIDTH / 2, TRACK_FRONT_Y + BIN_TOTAL_HT - 80,
           xcenter + BIN_WIDTH / 2 - 1, TRACK_FRONT_Y + BIN_TOTAL_HT - 1,
           makecol(0, 0, 0));
  for (int y = 0; y < 8; ++y) {
    int sliverY = TRACK_FRONT_Y + BIN_TOTAL_HT + y * 3;
    int c = trackpal[5][7 - y];
    rectfill(dst, xcenter - y * 2 - (2 + BIN_WIDTH / 2), sliverY, xcenter + (1 + BIN_WIDTH / 2) + y * 2, sliverY + 2, c);
  }
}

static const unsigned char tileColors[8][3] = {
  {3, 0, 0}, {3, 2, 0}, {3, 3, 0}, {0, 3, 0},
  {0, 3, 3}, {0, 1, 3}, {2, 0, 3}, {2, 2, 2}
};
static const unsigned char tileBlacklevel[5] = {0, 0, 0, 0, 102};
static const unsigned char tileGain[5] = {17*2, 17*3, 17*4, 17*5, 17*3};
void setTilePalette(int color) {
  for (int z = 0; z < 5; ++z) {
    tile_palette[z + 1].r = (tileBlacklevel[z]
                             + tileGain[z] * tileColors[color][0]) >> 2;
    tile_palette[z + 1].g = (tileBlacklevel[z]
                             + tileGain[z] * tileColors[color][1]) >> 2;
    tile_palette[z + 1].b = (tileBlacklevel[z]
                             + tileGain[z] * tileColors[color][2]) >> 2;
  }
}

static void drawTileOnTrack(BITMAP *dst, int xcenter, int track, int yoffset, fixed z, int color, bool circle) {
  setTilePalette(color);
  select_palette(tile_palette);
  clear_to_color(dtot_temp, bitmap_mask_color(dtot_temp));
  
  bool flip = false;
  int src_x = -track * 64;
  if (circle) {
    src_x = -192;
  }
  else if (track > 2) {
    src_x += 64;
    flip = true;
  }
  draw_sprite(dtot_temp, flip ? top_tiles_flip : top_tiles, src_x, 0);

  z += itofix(TRACK_DEPTH - 1);
  int dst_x = xcenter
              + fixdiv(TRACK_FRONT_WIDTH * TRACK_DEPTH * (track * 2 - N_TRACKS),
                       z * N_TRACKS * 2);
  int dst_y = fixdiv(TRACK_HEIGHT * 2 * TRACK_DEPTH, z);
  int dst_w = fixdiv(64 * TRACK_DEPTH, z);
  int dst_h = dst_w / 2;
  
  dst_y += TRACK_FRONT_Y - TRACK_HEIGHT * 2 - dst_h;
  if (flip) dst_x -= dst_w / 4;

  masked_stretch_blit(dtot_temp, dst,
                      0, 0, 64, 32,
                      dst_x, dst_y, dst_w, dst_h);
}

static void drawTilesOnTrack(const FKField *p) {
  FKFixed zOffset = 0x10000 - (p->progress & 0xFFFF);
  for (int z = TRACK_DEPTH - 1; z >= 0; --z) {
    if (p->trackTileColor[z] > 0) {
      FKFixed tileZ = (z << 16) + zOffset;
      int tileColor = p->trackTileColor[z] - 1;
      int tileX = p->trackTileX[z];
      bool circle = tileColor & 1;

      // wild tile colors
      if (tileColor >= 14) {
        tileColor = (ticks >> 3) & 0x07;
        circle = false;
      }

      drawTileOnTrack(back, SCREEN_W / 2, tileX, 0, tileZ, tileColor, circle);
    }
  }
}

static void drawTileOnPaddle(BITMAP *dst, int xcenter, int track, int y, int color, bool circle) {
  setTilePalette(color >= 0 ? color : 7);
  select_palette(tile_palette);
  clear_to_color(dtot_temp, bitmap_mask_color(dtot_temp));
  
  bool flip = false;
  int src_y = (color < 0) ? -32 : 0;
  int src_x = -track * 64;
  if (circle && color >= 0) {
    src_x = -192;
  }
  else if (track > 2) {
    src_x += 64;
    flip = true;
  }
  draw_sprite(dtot_temp, flip ? top_tiles_flip : top_tiles, src_x, src_y);

  fixed z = itofix(TRACK_DEPTH - 1);
  int dst_x = xcenter
              + fixdiv(TRACK_FRONT_WIDTH * TRACK_DEPTH * (track * 2 - N_TRACKS),
                       z * N_TRACKS * 2);
  int dst_y = fixdiv(TRACK_HEIGHT * 2 * TRACK_DEPTH, z);
  int dst_w = fixdiv(64 * TRACK_DEPTH, z);
  int dst_h = dst_w / 2;
  
  dst_y += 17 * y + TRACK_FRONT_Y - TRACK_HEIGHT * 2 - dst_h;
  if (flip) dst_x -= dst_w / 4;

  masked_stretch_blit(dtot_temp, dst,
                      0, 0, 64, 32,
                      dst_x, dst_y, dst_w, dst_h);
}

static void drawPaddle(FKField *p) {
  int h = PADDLE_HEIGHT;
  int dst_y = h;
  for (int y = 0; y < PADDLE_HEIGHT; ++y) {
    if (0 == p->paddleTileColor[y]) {
      h = y;
      dst_y = y + 1;
      break;
    }
  }

  drawTileOnPaddle(back, SCREEN_W / 2, p->paddleX, dst_y, -1, false);
  for (int y = h - 1; y >= 0; --y) {
    int tileColor = p->paddleTileColor[y] - 1;
    bool circle = tileColor & 1;

    // wild tile colors
    if (tileColor >= 14) {
      tileColor = (ticks >> 3) & 0x07;
      circle = false;
    }

    drawTileOnPaddle(back, SCREEN_W / 2, p->paddleX, --dst_y,
                     tileColor, circle);
  }

}

const char *const goalTypeTexts[FKGOAL_N_TYPES] = {
  "MARATHON",
  "SECONDS",
  "LINES",
  "DIAGONALS",
  "TILES",
  "POINTS",
  "HORIZONTALS"
};

static void drawToGo(const FKField *p) {
  int scoreTextOffset = (p->goalType == FKGOAL_POINTS) ? 112 : 96;
  int xcenter = SCREEN_W / 2;
  int white = makecol(255, 255, 255);
  int lightgray = makecol(191, 191, 191);
  int ledRed = makecol(255, 85, 85);
  textout_ex(back, picken_sans, goalTypeTexts[p->goalType],
             xcenter - scoreTextOffset, TRACK_FRONT_Y + 48, white, -1);
  if (p->goalType > 0) {
    int n = (p->goal > p->goalCount) ? p->goal - p->goalCount : 0;
    textout_ex(back, picken_sans, "TO GO",
               xcenter - scoreTextOffset, TRACK_FRONT_Y + 72, lightgray, -1);
    textprintf_right_ex(back, led_font,
                        xcenter + scoreTextOffset, TRACK_FRONT_Y + 56, ledRed, -1,
                        "%d", n);
  }
  textprintf_right_ex(back, picken_sans,
                      xcenter + 96, TRACK_FRONT_Y + 96, white, -1,
                      "%lu", (unsigned long)p->score);
  textout_ex(back, picken_sans, "SCORE:",
             xcenter - 96, TRACK_FRONT_Y + 96, white, -1);
}

void fkDrawIndicators(const FKField *p) {
  int xcenter = SCREEN_W / 2;
  
  select_palette(default_palette);
  for (int x = 0; x < p->lives; ++x) {
    draw_sprite(back, peacedes,
                xcenter - 96 + 40 * x, TRACK_FRONT_Y + 16);
  }
  if (p->lineClearFreezeTime) {
    int ledRed = makecol(255, 85, 85);
    textprintf_right_ex(back, led_font,
                        xcenter + 64, TRACK_FRONT_Y + 56, ledRed, -1,
                        "%lu", (unsigned long)p->lastLCPoints);
    textout_ex(back, picken_sans, "X",
               xcenter + 68, TRACK_FRONT_Y + 64, ledRed, -1);
    textprintf_ex(back, led_font,
                  xcenter + 80, TRACK_FRONT_Y + 56, ledRed, -1,
                  "%d", p->chain);
  } else {
    drawToGo(p);
  }
}

void eraseSprites(void) {
    // back of track
    blit(bg, back,
         SCREEN_W / 2 - 80, TRACK_FRONT_Y - TRACK_HEIGHT - 16,
         SCREEN_W / 2 - 80, TRACK_FRONT_Y - TRACK_HEIGHT - 16,
         160, TRACK_HEIGHT);
    // left side of track
    blit(bg, back,
         SCREEN_W / 2 - 160, TRACK_FRONT_Y - TRACK_HEIGHT,
         SCREEN_W / 2 - 160, TRACK_FRONT_Y - TRACK_HEIGHT,
         160 - TRACK_FRONT_WIDTH / 4, TRACK_HEIGHT);
    // right side of track
    blit(bg, back,
         SCREEN_W / 2 + TRACK_FRONT_WIDTH / 4, TRACK_FRONT_Y - TRACK_HEIGHT,
         SCREEN_W / 2 + TRACK_FRONT_WIDTH / 4, TRACK_FRONT_Y - TRACK_HEIGHT,
         160 - TRACK_FRONT_WIDTH / 4, TRACK_HEIGHT);
    // paddle and bin area
    blit(bg, back,
         SCREEN_W / 2 - 160, TRACK_FRONT_Y,
         SCREEN_W / 2 - 160, TRACK_FRONT_Y,
         320, BIN_TOTAL_HT + 24);
}

static void drawTileInBin(BITMAP *dst, int xcenter, int track, int y, int color, bool circle) {
  setTilePalette(color >= 0 ? color : 7);
  select_palette(tile_palette);
  clear_to_color(dtot_temp, bitmap_mask_color(dtot_temp));
  
  bool flip = false;
  int src_y = -64;
  int src_x = -track * 64;
  if (circle && color >= 0) {
    src_x = -192;
  }
  else if (track > 2) {
    src_x += 64;
    flip = true;
  }
  draw_sprite(dtot_temp, flip ? top_tiles_flip : top_tiles, src_x, src_y);

  int dst_x = xcenter
              + 28 * (track * 2 - N_TRACKS);
  int dst_y = TRACK_FRONT_Y + BIN_TOTAL_HT - 16 * y - 16;
  int dst_w = 64;
  int dst_h = 40;
  
  if (flip) dst_x -= dst_w / 8;

  masked_blit(dtot_temp, dst,
               0, 0, dst_x, dst_y,
               dst_w, dst_h);
}

static void drawTilesInBin(FKField *p) {
  for (int y = 0; y < BIN_HEIGHT; ++y) {
    for (int x1 = 0; x1 * 2 < N_TRACKS; ++x1) {
      int x = x1;
      for (int side = 0; side < 2; ++side) {
        if (p->binTileColor[y][x]) {
          int tile = p->binTileColor[y][x];
          int tileColor = (tile & 0x0F) - 1;
          bool circle = tileColor & 1;

          if (tile & ATTR_MARKED) {
            tileColor = ticks & 0x07;
          } else if (tileColor >= 14) {
          // wild tile colors
            tileColor = (ticks >> 3) & 0x07;
            circle = false;
          }

          drawTileInBin(back, SCREEN_W / 2,
                        x, y, tileColor, circle);
        }
        x = N_TRACKS - 1 - x;
        if (x == x1) break;
      } 
    }
  }
}

void fkDrawField(FKField *p) {
  eraseSprites();
  fkDrawIndicators(p);
  // tile spawn area
  debugSetColor(5);
  drawTrack(back, SCREEN_W / 2, false, p->progress >> 13);
  drawTilesInBin(p);
  if (p->paddleTileColor[PADDLE_HEIGHT - 1]) {
    // full paddle: draw it after the tiles on the track
    drawTilesOnTrack(p);
    drawPaddle(p);
  } else {
    drawPaddle(p);
    drawTilesOnTrack(p);
  }
  debugSetColor(6);
  vsync();
  scare_mouse();
  blit(back, screen,
       SCREEN_W / 2 - 160, 0,
       SCREEN_W / 2 - 160, 0,
       320, SCREEN_H);
  unscare_mouse();
  debugSetColor(7);
}

void drawSubtitle(const char *line1, const char *line2) {
  int white = makecol(255, 255, 255);
  textout_centre_ex(screen, picken_sans, line1, SCREEN_W / 2, 360, white, -1);
  textout_centre_ex(screen, picken_sans, line2, SCREEN_W / 2, 400, white, -1);
}

struct soundbalance_t {
  const char *name;
  unsigned short volume;
  unsigned short pitch;
};

static const struct soundbalance_t soundNames[] = {
  { "hold_wav",     128, 1000 },
  { "lock_wav",     128, 1000 },
  { "shift_wav",     64, 1000 },
  { "next2_wav",    128, 1000 },
  { "land_wav",     128, 1000 },
  { "line_wav",     128, 1000 },
  { "sectionup_wav",255, 1000 },
  { "gameover_wav", 255, 1000 }
};
void fkPlaySounds(FKBits sounds) {
  if (!sound_dat) return;
  for (int i = 0;
       sounds != 0 && i < sizeof(soundNames)/sizeof(soundNames[0]);
       ++i, sounds >>= 1) {
    if (sounds & 1) {
      const SAMPLE *spl = fdo(sound_dat, soundNames[i].name);
      if (spl) {
        play_sample(spl, soundNames[i].volume, 128, soundNames[i].pitch, false);
      }
    }
  }
}

void debugSetColor(int c) {
#if USE_DEBUG_SET_COLOR
  c = makecol(tileColors[c][0] * 85,
              tileColors[c][1] * 85,
              tileColors[c][2] * 85);
  rectfill(screen, 0, 0, 31, 31, c);
#endif
}

