/* Joystick binding configurator for Allegro programs

Copyright (C) 2009 Damian Yerrick

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "joyconfig.h"
#include "drawing.h"
#include <allegro.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>


/* RAW CONTROLLER HANDLING ****************************************/

FKBits buttonsHeld[MAX_JOYSTICKS];
FKBits lastButtonsHeld[MAX_JOYSTICKS];
FKBits buttonsPressed[MAX_JOYSTICKS];

uint32_t readAllPads(void) {
  FKBits changed = 0;

  poll_joystick();
  for(size_t j = 0; j < num_joysticks && j < MAX_JOYSTICKS; j++) {
    FKBits bits = 0;
    unsigned int bitPos = 0;

    lastButtonsHeld[j] = buttonsHeld[j];
    
    // read all axes
    for (size_t s = 0; s < joy[j].num_sticks; s++)
    {
      for (size_t x = 0; x < joy[j].stick[s].num_axis; x++)
      {
        if (joy[j].stick[s].axis[x].d1) {
          bits |= 1 << bitPos;
        } else if (joy[j].stick[s].axis[x].d2) {
          bits |= 2 << bitPos;
        }
        bitPos += 2;
      }
    }

    bitPos = 16;
    
    // read all buttons
    for(size_t b = 0; b < joy[j].num_buttons; b++)
    {
      int pressed = joy[j].button[b].b;
      if (pressed) {
        bits |= 1 << bitPos;
      }
      bitPos += 1;
    }
    buttonsHeld[j] = bits;

    // find any new presses
    bits &= ~lastButtonsHeld[j];
    if (bits != 0) {
      changed |= 1 << j;
    }
    buttonsPressed[j] = bits;
  }

  return changed;
}

/**
 * The signature for a controller is a 32-bit number made of 8 nibbles.
 * Nibbles 7 through 3 are the number of axes for sticks 0 through 4,
 * which Allegro thinks can be in the range 0 (stick don exits) to 3.
 * Nibbles 1 and 0 are the number of buttons.
 *
 * Notably, Allegro supports up to 15 total axes (3 on each of 5 sticks)
 * and 32 buttons, but joyconfig works with only the first 8 axes
 * and 16 buttons.
 */
FKBits getPadSig(int j) {
  if (j >= num_joysticks) {
    return 0;
  }
  FKBits sig = joy[j].num_buttons;
  for (size_t s = 0; s < joy[j].num_sticks; ++s) {
    sig |= joy[j].stick[s].num_axis << (4 * (7 - s));
  }
  return sig;
}

/* RAW TO VIRTUAL KEY TRANSLATION ********************************/

static signed short vkeyBindings[MAX_PLAYERS][MAX_VKEYS];

const char *const vkeyNames[MAX_VKEYS] = {
  "Menu up", "Menu down", "Menu OK",
  "Toss tile", "Speed up (whammy)", "Move left", "Move right", "Drop tile",
  "Fret 1", "Fret 2", "Fret 3", "Fret 4", "Fret 5"
};

unsigned int bfffo(FKBits rowBits) {
  unsigned int lineRow = 0;

  if (!rowBits) {
    return 32;
  }
  if ((rowBits & 0xFFFF) == 0) {
    rowBits >>= 16;
    lineRow += 16;
  }
  if ((rowBits & 0xFF) == 0) {
    rowBits >>= 8;
    lineRow += 8;
  }
  if ((rowBits & 0xF) == 0) {
    rowBits >>= 4;
    lineRow += 4;
  }
  if ((rowBits & 0x3) == 0) {
    rowBits >>= 2;
    lineRow += 2;
  }
  if ((rowBits & 0x1) == 0) {
    rowBits >>= 1;
    lineRow += 1;
  }
  return lineRow;
}

#if 0
void readAllPadsDemo(void) {
  int bgc = makecol(51, 102, 153);
  int fgc = makecol(255, 255, 255);
  while (!keypressed()) {
    FKBits changed = readAllPads();
    int player = bfffo(changed);

    for (size_t p = 0; p < MAX_JOYSTICKS && p < num_joysticks; ++p) {
      textprintf_ex(screen, font, 8, 10 * p + 10, fgc, bgc,
                    "player %d: held %08lx pressed %08lx",
                    p + 1, (unsigned long)buttonsHeld[p], (unsigned long)buttonsPressed[p]);
    }
    textprintf_ex(screen, font, 8, 10 * MAX_JOYSTICKS + 10, fgc, bgc,
                  "new presses on players %08lx namely%2d",
                  (unsigned long)changed, player);
    rest(50);
  }
  readkey();
}
#endif

#if 0
void getPadSigDemo(void) {
  int bgc = makecol(51, 102, 153);
  int fgc = makecol(255, 255, 255);

  clear_to_color(screen, bgc);
  for (size_t j = 0; j < MAX_JOYSTICKS && j < num_joysticks; ++j) {
    textprintf_ex(screen, font, 8, 10 * j + 10, fgc, bgc,
                  "controller %d sig is %08lx",
                  j + 1, (unsigned long)getPadSig(j));
  }
  readkey();
}
#endif

static int readVkeyBinding(unsigned int vkey) {
  if (vkey <= 0) {  // unassigned
    return 0;
  } else if (vkey < KEY_MAX) {  // keyboard
    return key[vkey];
  } else if (vkey < 256) {
    return 0;
  } else if (vkey < 256 * (MAX_JOYSTICKS + 1)) {  // joystick
    return buttonsHeld[(vkey >> 8) - 1] & (1 << (vkey & 0x1F));
  } else {
    return 0;
  }
}

FKBits readVkeys(unsigned int player) {
  FKBits held = 0;

  for (int a = 0; a < MAX_VKEYS; ++a) {
    int vkeyState = readVkeyBinding(vkeyBindings[player][a]);

    if (vkeyState) {
      held |= 1 << a;
    }
  }
  return held;
}

FKBits readMenuVkeys(void) {
  FKBits fixedKeys = 0;

  if (key[KEY_UP]) {
    fixedKeys = VKEY_MENUUP;
  }
  if (key[KEY_DOWN]) {
    fixedKeys |= VKEY_MENUDOWN;
  }
  if (key[KEY_ENTER]) {
    fixedKeys |= VKEY_MENUOK;
  }
  return fixedKeys ? fixedKeys : readVkeys(0);
}

/* CONFIGURATION DIALOG *****************************************/

enum {
  VKI_CONFIRM_1 = VKI_SPEED,
  VKI_CONFIRM_2 = VKI_DROP
};

int setupPlayerVkeys(unsigned int player) {
  signed short new_vkeys[MAX_VKEYS];
  char line1[128], line2[128];

  for (int a = 0; a < MAX_VKEYS; ++a) {

    // Ignore keys being held down when the function starts
    clear_keybuf();
    readAllPads();

    // Describe what user is supposed to do
    vsync();
    rectfill(screen, 0, 352, SCREEN_W, 447, makecol(51, 102, 153));
    usprintf(line2, "player %u %s", (unsigned int)player + 1, vkeyNames[a]);
    drawSubtitle("Press a button for", line2);

    for (bool beenSet = false; !beenSet; ) {
      if (keypressed()) {
        int scancode;
        ureadkey(&scancode);
        if (scancode == KEY_ESC) {
          return 0;
        } else if (scancode == KEY_TAB) {
          beenSet = true;
          new_vkeys[a] = 0;
        } else {
          beenSet = true;
          new_vkeys[a] = scancode;
        }
      }

      int players = readAllPads();
      if (players) {

        // find which player was in the bitfield
        int player = bfffo(players);
        FKBits bPressed = buttonsPressed[player];
        if (bPressed) {

          // Check buttons before directions
          int buttonBit = bfffo(bPressed >> 16);
          int bit = (bPressed >> 16)
                    ? (buttonBit + 16)
                    : bfffo(bPressed);

          new_vkeys[a] = ((player + 1) << 8) | bit;
          beenSet = true;
        }
      }
      rest(30);
    }

    // notify
    if (new_vkeys[a] == 0) {
      strcpy(line1, "Skipped assignment for");
      // line2 is still set
    } else if ((new_vkeys[a] >> 8) == 0) {
      strcpy(line1, "Pressed keyboard key");
      strcpy(line2, scancode_to_name(new_vkeys[a]));
    } else {
      usprintf(line1, "Pressed controller %d", new_vkeys[a] >> 8);
      if (new_vkeys[a] & 0x0010) {
        usprintf(line2, "button %d", (new_vkeys[a] & 0x000F) + 1);
      } else {
        usprintf(line2, "axis %d %c",
                 ((new_vkeys[a] & 0x000E) >> 1) + 1,
                 (new_vkeys[a] & 0x0001) ? '+' : '-');
      }
    }
    vsync();
    rectfill(screen, 0, 352, SCREEN_W, 447, makecol(51, 102, 153));
    drawSubtitle(line1, line2);
    rest(500);
  }

  // confirm the setting by pressing the first two vkeys
  vsync();
  rectfill(screen, 0, 352, SCREEN_W, 447, makecol(51, 102, 153));
  usprintf(line1, "Now press player %u", (unsigned int)player + 1);
  usprintf(line2, "%s + %s", vkeyNames[VKI_CONFIRM_1], vkeyNames[VKI_CONFIRM_2]);
  drawSubtitle(line1, line2);

  for (bool beenSet = false; !beenSet; ) {
    bool pressedAnything = false;

    pressedAnything = readAllPads();
    if (keypressed()) {
      int scancode;
      ureadkey(&scancode);
      if (scancode == KEY_ESC) {
        return 0;
      } else {
        pressedAnything = true;
      }
    }
    if (pressedAnything
        && readVkeyBinding(new_vkeys[VKI_CONFIRM_1])
        && readVkeyBinding(new_vkeys[VKI_CONFIRM_2])) {
      beenSet = true;
    }
  }

  for (int a = 0; a < MAX_VKEYS; ++a) {
    vkeyBindings[player][a] = new_vkeys[a];
  }
  vsync();
  rectfill(screen, 0, 352, SCREEN_W, 447, makecol(51, 102, 153));
  usprintf(line1, "Player %u's buttons", (unsigned int)player + 1);
  drawSubtitle(line1, "have been assigned.");
  rest(500);

  return 1;
}

void demoReadVkeys(void) {
  int lastVkeys[MAX_PLAYERS] = {0};
  int vkeys[MAX_PLAYERS] = {0};
  int white = makecol(255, 255, 255);

  clear_bitmap(screen);
  textout_ex(screen, picken_sans, "Input Test (Esc to end)", 48, 48, white, -1);

  while (1) {
    bool changed = false;

    if (keypressed()) {
      int scancode;
      ureadkey(&scancode);
      if (scancode == KEY_ESC) {
        break;
      }
    }
    readAllPads();
    for (int p = 0; p < MAX_PLAYERS; ++p) {
      lastVkeys[p] = vkeys[p];
      vkeys[p] = readVkeys(p);
      if (lastVkeys[p] != vkeys[p]) {
        changed = true;
      }
    }
    if (changed) {
      vsync();
      rectfill(screen, 0, 96, SCREEN_W - 1, 431, 0);
      int y = 96, x = 64;
      for (unsigned int p = 0; p < MAX_PLAYERS && x <= 400; ++p) {
        for (unsigned int a = 0; a < MAX_VKEYS && x <= 400; ++a) {
          if (vkeys[p] & (1 << a)) {
            textprintf_ex(screen, picken_sans, x, y, white, -1,
                          "p%u %s",
                          (unsigned int)p + 1, vkeyNames[a]);
            y += 24;
            if (y > 384) {
              y = 96;
              x += 336;
            }
          }
        }
      }
    }
    rest(30);
  }
}

/* LOADING AND SAVING VIRTUAL KEY CONFIGURATION ***********************/

static const char keysFilename[] = "fk_keys.ini";

bool saveVkeys(void) {
  FILE *fp = fopen(keysFilename, "wt");
  if (!fp) {
    return false;
  }
  fprintf(fp, "sig=");
  for (unsigned int j = 0; j < num_joysticks; ++j) {
    fprintf(fp, "%08lx", (unsigned long int)getPadSig(j));
  }
  fputc('\n', fp);
  for (unsigned int p = 0; p < MAX_PLAYERS; ++p) {
    fprintf(fp, "%u=", p + 1);
    for (unsigned int a = 0; a < MAX_VKEYS; ++a) {
      fprintf(fp, "%04x", vkeyBindings[p][a]);
    }
    fputc('\n', fp);
  }
  fclose(fp);
  return true;
}

static bool read_xdigits(const char *txt, size_t length, FKBits *out) {
  FKBits data = 0;
  for(; length > 0; --length) {
    int ch = *txt++;
    unsigned int val = 0;
    if (ch >= '0' && ch <= '9') {
      val = ch - '0';
    } else if (ch >= 'A' && ch <= 'F') {
      val = ch + 10 - 'A';
    } else if (ch >= 'a' && ch <= 'f') {
      val = ch + 10 - 'a';
    } else {
      return false;
    }
    data = (data << 4) | val;
  }
  if (out) {
    *out = data;
  }
  return true;
}

/**
 * Compares a string containing 8-digit hex controller signatures
 * to the controllers currently plugged in.
 * @return false
 */
static bool samePadSigs(const char *oldSig) {
  size_t sigsLen = strlen(oldSig);

  // strip trailing whitespace
  while (sigsLen > 0
         && isspace(oldSig[sigsLen - 1])) {
    --sigsLen;
  }

  // If the number of controllers has changed,
  // the signature is invalid.
  if (sigsLen != 8 * num_joysticks) {
    return false;
  }

  for (unsigned int j = 0;
       j < num_joysticks;
       ++j, oldSig += 8) {
    FKBits sig = 0;
    if (!read_xdigits(oldSig, 8, &sig)) {
      return false;
    }
    if (sig != getPadSig(j)) {
      return false;
    }
  }
  return true;
}

static bool loadPlayerVkeys(size_t player, const char *bindings) {
  size_t bindingsLen = strlen(bindings);
//  char trace[128];

  // strip trailing whitespace
  while (bindingsLen > 0
         && isspace(bindings[bindingsLen - 1])) {
    --bindingsLen;
  }
  if (bindingsLen != 4 * MAX_VKEYS) {
    return false;
  }
//  usprintf(trace, "len: %u", (unsigned int)bindingsLen);
//  alert("loadPlayerVkeys", bindings, trace, "OK", 0, 13, 0);

  for (unsigned int a = 0;
       a < MAX_VKEYS;
       ++a, bindings += 4) {
    FKBits binding = 0;
    if (!read_xdigits(bindings, 4, &binding)) {
      return false;
    }
    vkeyBindings[player][a] = binding;
//    usprintf(trace, "p%u a%u = %04x",
//             (unsigned int)player, a, (unsigned int)binding);
//    alert("loadPlayerVkeys", bindings, trace, "OK", 0, 13, 0);
  }

  return true;
}

void clearVkeys(void) {
  for (int p = 0; p < MAX_PLAYERS; ++p) {
    for (int a = 0; a < MAX_VKEYS; ++a) {
      vkeyBindings[p][a] = 0;
    }
  }
}

bool loadVkeys(void) {
  char line[256];

  FILE *fp = fopen(keysFilename, "rt");
  if (!fp) {
    return false;
  }

  while (fgets(line, sizeof(line), fp)) {
    if (!strncmp("sig=", line, 4)) {

      // The rest of the line is controller signatures.
      // If they don't match, don't load the file.
      if (!samePadSigs(line + 4)) {
        return false;
      }
    } else if (line[0] >= '1' && line[0] <= '8' && line[1] == '=') {

      // These are the vkey bindings for one player, four digits each.
      if (!loadPlayerVkeys(line[0] - '1', line + 2)) {
        return false;
      }
    }
  }

  fclose(fp);
  return true;
}
