/* FK: tile matching game

Copyright (C) 2009 Damian Yerrick

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "fk.h"

static void findHorz(FKField *p) {
  p->lastlc.horizontals[0] = 0;
  p->lastlc.horizontals[1] = 0;
  p->lastlc.horizontals[2] = 0;

  for (int y = 0; y < BIN_HEIGHT; ++y) {
    int left = 0;
    while (left + 2 < N_TRACKS) {
      int x = left + 1;
      int nextLeft = x;
      int matchColor = p->binTileColor[y][left] & 0x0F;
      if (matchColor == 0) {
        // Skip empty space
        left = nextLeft;
        continue;
      }
      while (x < N_TRACKS) {
        int blkHere = p->binTileColor[y][x] & 0x0F;
        if (blkHere == 0) {
          // empty space
          nextLeft = x + 1;
          break;
        } else if (blkHere == WILD_COLOR) {
          x += 1;
        } else if (matchColor == WILD_COLOR || matchColor == blkHere) {
          // the trick to finding overlapping wild-tile matches is to
          // start the next search at the block after the nonwild tile
          nextLeft = x + 1;
          x += 1;
          matchColor = blkHere;
        } else {
          break;
        }
      }
      int matchLen = x - left;
      if (matchLen >= 3) {
        p->lastlc.horizontals[matchLen - 3] += 1;
        for (int x1 = left; x1 < x; ++x1) {
          p->binTileColor[y][x1] |= ATTR_MARKED;
        }
      }
      left = nextLeft;
    }
  }
}

static void findVert(FKField *p) {
  p->lastlc.verticals[0] = 0;
  p->lastlc.verticals[1] = 0;
  p->lastlc.verticals[2] = 0;

  for (int x = 0; x < N_TRACKS; ++x) {
    int bottom = 0;
    while (bottom + 2 < BIN_HEIGHT) {
      int y = bottom + 1;
      int nextBottom = y;
      int matchColor = p->binTileColor[bottom][x] & 0x0F;
      if (matchColor == 0) {
        // Skip empty space
        bottom = nextBottom;
        continue;
      }
      while (x < N_TRACKS) {
        int blkHere = p->binTileColor[y][x] & 0x0F;
        if (blkHere == 0) {
          // empty space
          nextBottom = y + 1;
          break;
        } else if (blkHere == WILD_COLOR) {
          y += 1;
        } else if (matchColor == WILD_COLOR || matchColor == blkHere) {
          // the trick to finding overlapping wild-tile matches is to
          // start the next search at the block after the nonwild tile
          nextBottom = y + 1;
          y += 1;
          matchColor = blkHere;
        } else {
          break;
        }
      }
      int matchLen = y - bottom;
      if (matchLen >= 3) {
        p->lastlc.verticals[matchLen - 3] += 1;
        for (int y1 = bottom; y1 < y; ++y1) {
          p->binTileColor[y1][x] |= ATTR_MARKED;
        }
      }
      bottom = nextBottom;
    }
  }
}

static void findUpDiag(FKField *p) {
  p->lastlc.diagonals[0] = 0;
  p->lastlc.diagonals[1] = 0;
  p->lastlc.diagonals[2] = 0;

  for (int y = -2; y < BIN_HEIGHT - 2; ++y) {
    int left = 0;
    if (y < 0) {
      left = -y;
    }
    while (left + 2 < N_TRACKS && left + y < BIN_HEIGHT) {
      int x = left + 1;
      int nextLeft = x;
      int matchColor = p->binTileColor[y + left][left] & 0x0F;
      if (matchColor == 0) {
        // Skip empty space
        left = nextLeft;
        continue;
      }
      while (x < N_TRACKS && x + y < BIN_HEIGHT) {
        int blkHere = p->binTileColor[y + x][x] & 0x0F;
        if (blkHere == 0) {
          // empty space
          nextLeft = x + 1;
          break;
        } else if (blkHere == WILD_COLOR) {
          x += 1;
        } else if (matchColor == WILD_COLOR || matchColor == blkHere) {
          // the trick to finding overlapping wild-tile matches is to
          // start the next search at the block after the nonwild tile
          nextLeft = x + 1;
          x += 1;
          matchColor = blkHere;
        } else {
          break;
        }
      }
      int matchLen = x - left;
      if (matchLen >= 3) {
        p->lastlc.diagonals[matchLen - 3] += 1;
        for (int x1 = left; x1 < x; ++x1) {
          p->binTileColor[y + x1][x1] |= ATTR_MARKED;
        }
      }
      left = nextLeft;
    }
  }
}

static void findDownDiag(FKField *p) {
  // don't clear out counts because updiag already did

  for (int y = 2; y < BIN_HEIGHT + 2; ++y) {
    int left = 0;
    if (y > BIN_HEIGHT - 1) {
      left = y - (BIN_HEIGHT - 1);
    }
    while (left + 2 < N_TRACKS && y - left >= 2) {
      int x = left + 1;
      int nextLeft = x;
      int matchColor = p->binTileColor[y - left][left] & 0x0F;
      if (matchColor == 0) {
        // Skip empty space
        left = nextLeft;
        continue;
      }
      while (x < N_TRACKS && y - x >= 0) {
        int blkHere = p->binTileColor[y - x][x] & 0x0F;
        if (blkHere == 0) {
          // empty space
          nextLeft = x + 1;
          break;
        } else if (blkHere == WILD_COLOR) {
          x += 1;
        } else if (matchColor == WILD_COLOR || matchColor == blkHere) {
          // the trick to finding overlapping wild-tile matches is to
          // start the next search at the block after the nonwild tile
          nextLeft = x + 1;
          x += 1;
          matchColor = blkHere;
        } else {
          break;
        }
      }
      int matchLen = x - left;
      if (matchLen >= 3) {
        p->lastlc.diagonals[matchLen - 3] += 1;
        for (int x1 = left; x1 < x; ++x1) {
          p->binTileColor[y - x1][x1] |= ATTR_MARKED;
        }
      }
      left = nextLeft;
    }
  }
}

void fkLineClear(FKField *p) {
  for (int x = 0; x < N_TRACKS; ++x) {
    for (int y = 0; y < BIN_HEIGHT; ++y) {
      if (ATTR_MARKED & p->binTileColor[y][x]) {
        p->binTileColor[y][x] = 0;
      }
    }
    
    int dsty = 0;
    for (int y = 0; y < BIN_HEIGHT; ++y) {
      int srcTile = p->binTileColor[y][x];
      if (srcTile) {
        if (dsty < y) {
          p->binTileColor[y][x] = 0;
          p->binTileColor[dsty][x] = srcTile;
          p->sounds |= FKSND_DROP;
        }
        ++dsty;
      }
    }
  }
  p->needLineCheck = true;
}

void fkLineCheck(FKField *p) {
  int nClears = 0;

  p->needLineCheck = false;
  findHorz(p);
  nClears += p->lastlc.horizontals[0]
             + p->lastlc.horizontals[1]
             + p->lastlc.horizontals[2];
  findVert(p);
  nClears += p->lastlc.verticals[0]
             + p->lastlc.verticals[1]
             + p->lastlc.verticals[2];
  findUpDiag(p);
  findDownDiag(p);
  nClears += p->lastlc.diagonals[0]
             + p->lastlc.diagonals[1]
             + p->lastlc.diagonals[2];
  
  if (nClears > 0) {
    p->sounds |= FKSND_LINE;
    p->chain += nClears;
    uint32_t pointsNow = 0;
    p->lineClearFreezeTime = 60;
    
    // this won't ever get taken, but it's there because a future
    // update might include 0-lcd mode
    if (p->lineClearFreezeTime == 0) {
      fkLineClear(p);
    }
    pointsNow +=  1000 * p->lastlc.horizontals[0];
    pointsNow +=  5000 * p->lastlc.horizontals[1];
    pointsNow += 10000 * p->lastlc.horizontals[2];
    pointsNow +=    50 * p->lastlc.verticals[0];
    pointsNow += 10000 * p->lastlc.verticals[1];
    pointsNow += 15000 * p->lastlc.verticals[2];
    pointsNow +=  5000 * p->lastlc.diagonals[0];
    pointsNow += 10000 * p->lastlc.diagonals[1];
    pointsNow += 20000 * p->lastlc.diagonals[2];
    p->lastLCPoints = pointsNow;
    fkAddPoints(p, pointsNow * p->chain);
    
    int linesH = p->lastlc.horizontals[0]
                 + 2 * p->lastlc.horizontals[1]
                 + 3 * p->lastlc.horizontals[2];
    int linesV = p->lastlc.diagonals[0]
                 + 2 * p->lastlc.diagonals[1]
                 + 3 * p->lastlc.diagonals[2];
    int linesD = p->lastlc.diagonals[0]
                 + 2 * p->lastlc.diagonals[1]
                 + 3 * p->lastlc.diagonals[2];
    if (p->goalType == FKGOAL_LINES) {
      p->goalCount += linesH;
    }
    if (p->goalType == FKGOAL_LINES || p->goalType == FKGOAL_HORIZONTALS) {
      p->goalCount += linesV;
    }
    if (p->goalType == FKGOAL_LINES || p->goalType == FKGOAL_DIAGONALS) {
      p->goalCount += linesD;
    }
  } else {
    p->chain = 0;
  }
}

