/* FK: tile matching game

Copyright (C) 2009 Damian Yerrick

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "drawing.h"
#include "joyconfig.h"


void initTrack(FKField *p) {
  for (int z = 0; z < TRACK_DEPTH; ++z) {
    p->trackTileColor[z] = 0;
  }
}

void initPaddle(FKField *p) {
  p->paddleX = (N_TRACKS - 1) / 2;
  p->paddleTileColor[0] = 1;
  for (size_t y = 1; y < PADDLE_HEIGHT; ++y) {
    p->paddleTileColor[y] = 0;
  }
}

void initBin(FKField *p) {
  for (size_t y = 0; y < BIN_HEIGHT; ++y) {
    for (size_t x = 0; x < N_TRACKS; ++x) {
      p->binTileColor[y][x] = 0;
    }
  }
  p->needLineCheck = false;
  p->lineClearFreezeTime = 0;
  p->chain = 0;
}

unsigned int fkRand(FKField *p) {
  p->seed = p->seed * 2147001325 + 715136305;
  return p->seed >> 17;
}

void fkSrand(FKField *p, uint32_t value) {
  p->seed = value;
}

/**
 * Sets up the randomizer.
 *
 * Reviewers of Klax games have complained about the game not giving
 * the right color for a long time.  In fact, one reviewer of the NES
 * version of Klax accused the game of using the Bastet randomizer:
 * http://www.nesplayer.com/reviews/klax.htm
 * But it's probably the same observation bias that led some players
 * of older Tetris games, which used an uncorrelated uniform
 * randomizer, to accuse those games of doing a Bastet too.
 *
 * So instead, FK uses a randomizer reminiscent of Tetris the Grand
 * Master or LJ65.  It keeps a list of which tile colors are most and
 * least recently used, and it always picks one of the four least
 * recently used tiles. 
 */
void init_generateTile(FKField *p, int nColors) {
  if (nColors > N_COLORS + 1) nColors = N_COLORS + 1;
  for (int i = 0; i < nColors; ++i) {
    p->history[i] = i;
  }
  p->nColors = nColors;
  p->density = 12;
  p->densityCounter = 255;
}

/**
 * Chooses a random color from among the four least recently used,
 * and puts a tile of that color on the farthest row of the track,
 * but not on the same X coordinate as the last tile.
 */
void generateTile(FKField *p) {
  // get a random X position (0-3: green, red, yellow, blue)
  // and a random depth into the history (0-3)
  int r = fkRand(p);
  int nextX = (r >> 2) & 0x03;
  int nextDepth = r & 0x03;

  // If the X position matches that of the last tile,
  // use 4 (orange) instead.
  if (nextX == p->trackTileX[TRACK_DEPTH - 1]) nextX = 4;

  // LRU color randomizer 
  int nextColor = p->history[nextDepth];
  for (int i = nextDepth; i < p->nColors - 1; ++i) {
    p->history[i] = p->history[i + 1];
  }
  p->history[p->nColors - 1] = nextColor;

  p->trackTileColor[TRACK_DEPTH - 1] = nextColor + 1;
  p->trackTileX[TRACK_DEPTH - 1] = nextX;
}

static bool tryCatching(FKField *p) {

  // the paddle must be there...
  if (p->trackTileX[0] != p->paddleX) return false;
  
  // and not already full.
  if (p->paddleTileColor[PADDLE_HEIGHT - 1] != 0) return false;

  for (int y = PADDLE_HEIGHT - 2; y >= 0; --y) {
    p->paddleTileColor[y + 1] = p->paddleTileColor[y];
  }
  p->paddleTileColor[0] = p->trackTileColor[0];
  
  // rewards
  fkAddPoints(p, 5);
  if (p->goalType == FKGOAL_TILES) {
    ++p->goalCount;
  }
  p->sounds |= FKSND_CATCH;
  return true;
}

void tilesForwardBy1(FKField *p) {
  if (p->trackTileColor[0]) {
    if (!tryCatching(p)) {
      if (p->lives > 0) {
        --p->lives;
        p->sounds |= FKSND_MISS;
      }
    }
  }

  // move tiles forward by 1
  for (int z = 0; z < TRACK_DEPTH - 1; ++z) {
    p->trackTileX[z] = p->trackTileX[z + 1];
    p->trackTileColor[z] = p->trackTileColor[z + 1];
  }

  int countup = p->density + p->densityCounter;
  if (countup >= 256) {
    p->densityCounter = countup - 256;
    if (p->density < 80) {
      ++p->density;
    }
    generateTile(p);
  } else {
    p->densityCounter = countup;
    p->trackTileColor[TRACK_DEPTH - 1] = 0;
  }
}

void tilesForwardTo(FKField *p, FKFixed new_z) {
  while (p->progress >> 16 < new_z >> 16) {
    tilesForwardBy1(p);
    p->progress += 0x10000;
  }
  p->progress = new_z;
}


void fkAddPoints(FKField *p, uint32_t nPoints) {
  p->score += nPoints;
  if (p->goalType == FKGOAL_POINTS) {
    p->goalCount += nPoints;
  }
}

static void shiftPaddle(FKField *p, FKControl *c) {
  int dasKeys = c->newKeys;
  int oldPaddleX = p->paddleX;
  
  if (c->newKeys & VKEY_LEFT) {
    c->dasDir = VKEY_LEFT;
    c->dasCharge = 0;
  } else if (c->newKeys & VKEY_RIGHT) {
    c->dasDir = VKEY_RIGHT;
    c->dasCharge = 0;
  }
  if (c->lastKeys & c->dasDir) {
    ++c->dasCharge;
    if (c->dasCharge >= 15) {
      c->dasCharge -= 2;
      dasKeys |= c->dasDir;
    }
  } else {
    c->dasDir = 0;
  }
  if (p->paddleX > 0 && (dasKeys & VKEY_LEFT)) {
    --p->paddleX;
  }
  if (p->paddleX < N_TRACKS - 1 && (dasKeys & VKEY_RIGHT)) {
    ++p->paddleX;
  }

  int m_track = (mouse_x + 128 - SCREEN_W / 2) * N_TRACKS / 256;
  if (m_track < 0) {
    m_track = 0;
  } else if (m_track > N_TRACKS - 1) {
    m_track = N_TRACKS - 1;
  }
  if (m_track != c->mouseLastTrack) {
    c->mouseLastTrack = m_track;
    p->paddleX = m_track;
  }
  for (int x = 0; x < N_TRACKS; ++x) {
    int fret = x;
    if (VKEY_FRET(fret) & c->newKeys) {
      p->paddleX = fret;
    }
  }
  if (p->paddleX != oldPaddleX) {
    p->sounds |= FKSND_SHIFT;
  }
}

static unsigned int popTileFromPaddle(FKField *p) {
  int popped = p->paddleTileColor[0];
  if (!popped) {
    return 0;
  }
  for (int y1 = 0; y1 < PADDLE_HEIGHT - 1; ++y1) {
    p->paddleTileColor[y1] = p->paddleTileColor[y1 + 1];
  }
  p->paddleTileColor[PADDLE_HEIGHT - 1] = 0;
  return popped;
}

static void dropTileInBin(FKField *p) {
  if (p->binTileColor[BIN_HEIGHT - 1][p->paddleX]) return;
  int popped = popTileFromPaddle(p);
  if (!popped) return;
  
  for (int y = 0; y < BIN_HEIGHT; ++y) {
    if (!p->binTileColor[y][p->paddleX]) {
      p->binTileColor[y][p->paddleX] = popped;
      p->sounds |= FKSND_DROP;
      if (p->lineClearFreezeTime == 0) {
        p->needLineCheck = true;
      }
      return;
    }
  }
}

static void tossTileToTrack(FKField *p) {
  for (int z = TOSS_DEPTH; z < TRACK_DEPTH; ++z) {
    if (!p->trackTileColor[z]) {
      int popped = popTileFromPaddle(p);
      if (popped) {
        p->trackTileColor[z] = popped;
        p->trackTileX[z] = p->paddleX;
        p->sounds |= FKSND_TOSS;
      }
      return;
    }
  }
}

void fkFrame(FKField *p, FKControl *c) {
  //FKBits lastFrameSounds = p->sounds;
  p->sounds = 0;

  debugSetColor(0);
  readAllPads();
  FKBits j = readVkeys(0);
  if (mouse_b & 0x01) {
    j |= VKEY_DROP;
  }
  if (mouse_b & 0x02) {
    j |= VKEY_TOSS;
  }
  c->newKeys = j & ~c->lastKeys;
  c->lastKeys = j;

  debugSetColor(1);
  shiftPaddle(p, c);
  if (VKEY_DROP & c->newKeys) {
    dropTileInBin(p);
  }
  debugSetColor(2);
  if (VKEY_TOSS & c->newKeys) {
    tossTileToTrack(p);
  }
  debugSetColor(3);
  if (p->needLineCheck) {
    fkLineCheck(p);
  }
  debugSetColor(4);
  if (p->lineClearFreezeTime > 0) {
    if (--p->lineClearFreezeTime == 0) {
      fkLineClear(p);
    }
  } else {
    FKFixed speed = 0x0c00;
    if (VKEY_SPEED & c->lastKeys) {
      speed <<= 2;
    }
    tilesForwardTo(p, p->progress + speed);
    if (p->lives == 0) {
      p->sounds |= FKSND_GAMEOVER;
    }
  }
  ++p->waveTime;
}


void play(FKField *p) {
  FKControl c1 = {~0};
  init_generateTile(p, 9);
  p->history[8] = 14;
  initTrack(p);
  initPaddle(p);
  initBin(p);

  p->lives = 5;
  p->score = 0;
  p->goalType = FKGOAL_MARATHON;
  p->goal = 0;
  p->goalCount = 0;

  int lastTicks = ticks;
  bool done = false;
  while (!done) {
    fkFrame(p, &c1);
    if (ticks <= lastTicks) {
      fkDrawField(p);
      while (ticks <= lastTicks) {
        rest(5);
      }
    }
    ++lastTicks;
    
    if (keypressed()) {
      int sc;
      ureadkey(&sc);
      if (sc == KEY_ESC) {
        done = true;
      }
    }
    if (p->lives == 0) {
      done = true;
    }
    if (p->goalType > 0
        && p->goalCount >= p->goal
        && p->lineClearFreezeTime == 0) {
      done = true;
      p->sounds |= FKSND_SECTIONUP;
    }
    fkPlaySounds(p->sounds);
  }
}

